// -*- c++ -*-

/*

 Copyright 2019-2020 Alain Dargelas

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

/*
 * File:   ElaboratorListener.h
 * Author: alaindargelas
 *
 * Created on May 6, 2020, 10:03 PM
 */

#include <map>
#include <stack>
#include <iostream>

#include "headers/VpiListener.h"
#include "headers/clone_tree.h"

namespace UHDM {
  
class ElaboratorListener : public VpiListener {
  
public:
  
  ElaboratorListener (Serializer* serializer, bool debug = false) : serializer_(serializer), debug_(debug) {}

  // Bind to a net in the current instance
  net* bindNet(const std::string& name) {
    if (instStack_.size()) {
      ComponentMap& netMap = instStack_.top().second.first;
      ComponentMap::iterator netItr = netMap.find(name);
      if (netItr != netMap.end()) {
        return (net*) (*netItr).second;
      }
    }
    return nullptr;
  }

  // Bind to a net or parameter in the current instance
  any* bindAny(const std::string& name) {
    if (instStack_.size()) {
      ComponentMap& netMap = instStack_.top().second.first;
      ComponentMap::iterator netItr = netMap.find(name);
      if (netItr != netMap.end()) {
        return (any*) (*netItr).second;
      }
  
      ComponentMap& paramMap = instStack_.top().second.second;
      ComponentMap::iterator paramItr = paramMap.find(name);
      if (paramItr != paramMap.end()) {
        return (any*) (*paramItr).second;
      }
    }
    return nullptr;
  }

  any* bindParam(const std::string& name) {
    if (instStack_.size()) {
      ComponentMap& paramMap = instStack_.top().second.second;
      ComponentMap::iterator paramItr = paramMap.find(name);
      if (paramItr != paramMap.end()) {
        return (any*) (*paramItr).second;
      }
    }
    return nullptr;
  }
  
protected:
  typedef std::map<std::string, const BaseClass*> ComponentMap;

  void enterModule(const module* object, const BaseClass* parent,
                   vpiHandle handle, vpiHandle parentHandle) override {
    module* inst = (module*) object;
    bool topLevelModule         = object->VpiTopModule();
    const std::string& instName = object->VpiName(); 
    const std::string& defName  = object->VpiDefName(); 
    bool flatModule             = (instName == "") && ((object->VpiParent() == 0) ||
                                                       ((object->VpiParent() != 0) && (object->VpiParent()->VpiType() != vpiModule)));
                                  // false when it is a module in a hierachy tree
    if (debug_)
      std::cout << "Module: " << defName << " (" << instName << ") Flat:"  << flatModule << ", Top:" << topLevelModule << std::endl;

    if (flatModule) {
      // Flat list of module (unelaborated)
      flatComponentMap_.insert(std::make_pair(object->VpiDefName(), object));
    } else {
      // Hierachical module list (elaborated)

      // Collect instance elaborated nets
      ComponentMap netMap;
      if (object->Nets()) {
        for (net* net : *object->Nets()) {
          netMap.insert(std::make_pair(net->VpiName(), net));
        }
      }

      // Collect instance parameters, defparams 
      ComponentMap paramMap;
      if (object->Parameters()) {
        for (any* param : *object->Parameters()) {
          paramMap.insert(std::make_pair(param->VpiName(), param));
        }
      }
      if (object->Def_params()) {
        for (def_param* param : *object->Def_params()) {
          paramMap.insert(std::make_pair(param->VpiName(), param));
        }
      }
      if (object->Variables()) {
        for (variables* var : *object->Variables()) {
          paramMap.insert(std::make_pair(var->VpiName(), var));
        }
      }

      // Check if Module instance has a definition, collect enums
      ComponentMap::iterator itrDef = flatComponentMap_.find(defName);      
      if (itrDef != flatComponentMap_.end()) {
        const BaseClass* comp = (*itrDef).second;
        int compType = comp->VpiType();
        switch (compType) {
          case vpiModule: {
            module* defMod = (module*) comp;
            if (defMod->Typespecs()) {
              for (typespec* tps : *defMod->Typespecs()) {
                if (tps->UhdmType() == uhdmenum_typespec) {
                  enum_typespec* etps = (enum_typespec*)tps;
                  for (enum_const* econst : * etps->Enum_consts()) {
                    paramMap.insert(std::make_pair(econst->VpiName(), econst));
                  }
                }
              }
            }
          }
        }
      }
       
      // Push instance context on the stack
      instStack_.push(std::make_pair(object, std::make_pair(netMap, paramMap)));
      
      // Check if Module instance has a definition
      if (itrDef != flatComponentMap_.end()) {
        const BaseClass* comp = (*itrDef).second;
        int compType = comp->VpiType();
        switch (compType) {
        case vpiModule: {
          module* defMod = (module*) comp;

          if (auto obj = defMod->Instance_array()) {
            inst->Instance_array(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto vec = defMod->Process()) {
            auto clone_vec = serializer_->MakeProcess_stmtVec();
            inst->Process(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Primitives()) {
            auto clone_vec = serializer_->MakePrimitiveVec();
            inst->Primitives(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Primitive_arrays()) {
            auto clone_vec = serializer_->MakePrimitive_arrayVec();
            inst->Primitive_arrays(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto obj = defMod->Global_clocking()) {
            inst->Global_clocking(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto obj = defMod->Default_clocking()) {
            inst->Default_clocking(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto obj = defMod->Module_array()) {
            inst->Module_array(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto vec = defMod->Interfaces()) {
            auto clone_vec = serializer_->MakeInterfaceVec();
            inst->Interfaces(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Interface_arrays()) {
            auto clone_vec = serializer_->MakeInterface_arrayVec();
            inst->Interface_arrays(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Cont_assigns()) {
            auto clone_vec = serializer_->MakeCont_assignVec();
            inst->Cont_assigns(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Modules()) {
            auto clone_vec = serializer_->MakeModuleVec();
            inst->Modules(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Module_arrays()) {
            auto clone_vec = serializer_->MakeModule_arrayVec();
            inst->Module_arrays(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Mod_paths()) {
            auto clone_vec = serializer_->MakeMod_pathVec();
            inst->Mod_paths(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Tchks()) {
            auto clone_vec = serializer_->MakeTchkVec();
            inst->Tchks(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Def_params()) {
            auto clone_vec = serializer_->MakeDef_paramVec();
            inst->Def_params(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Io_decls()) {
            auto clone_vec = serializer_->MakeIo_declVec();
            inst->Io_decls(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Alias_stmts()) {
            auto clone_vec = serializer_->MakeAlias_stmtVec();
            inst->Alias_stmts(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Clocking_blocks()) {
            auto clone_vec = serializer_->MakeClocking_blockVec();
            inst->Clocking_blocks(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Gen_scope_arrays()) {
            auto clone_vec = serializer_->MakeGen_scope_arrayVec();
            inst->Gen_scope_arrays(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto obj = defMod->Expr_dist()) {
            inst->Expr_dist(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto vec = defMod->Task_funcs()) {
            auto clone_vec = serializer_->MakeTask_funcVec();
            inst->Task_funcs(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto obj = defMod->Instance()) {
            inst->Instance(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto vec = defMod->Programs()) {
            auto clone_vec = serializer_->MakeProgramVec();
            inst->Programs(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Program_arrays()) {
            auto clone_vec = serializer_->MakeProgramVec();
            inst->Program_arrays(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Array_nets()) {
            auto clone_vec = serializer_->MakeArray_netVec();
            inst->Array_nets(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Spec_params()) {
            auto clone_vec = serializer_->MakeSpec_paramVec();
            inst->Spec_params(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Class_defns()) {
            auto clone_vec = serializer_->MakeClass_defnVec();
            inst->Class_defns(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto obj = defMod->Module()) {
            inst->Module(obj->DeepClone(serializer_, this, defMod));
          }
          if (auto vec = defMod->Assertions()) {
            auto clone_vec = serializer_->MakeAnyVec();
            inst->Assertions(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Concurrent_assertions()) {
            auto clone_vec = serializer_->MakeConcurrent_assertionsVec();
            inst->Concurrent_assertions(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Variables()) {
            auto clone_vec = serializer_->MakeVariablesVec();
            inst->Variables(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Scopes()) {
            auto clone_vec = serializer_->MakeScopeVec();
            inst->Scopes(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Typespecs()) {
            auto clone_vec = serializer_->MakeTypespecVec();
            inst->Typespecs(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Property_decls()) {
            auto clone_vec = serializer_->MakeProperty_declVec();
            inst->Property_decls(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Sequence_decls()) {
            auto clone_vec = serializer_->MakeSequence_declVec();
            inst->Sequence_decls(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Named_events()) {
            auto clone_vec = serializer_->MakeNamed_eventVec();
            inst->Named_events(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Named_event_arrays()) {
            auto clone_vec = serializer_->MakeNamed_event_arrayVec();
            inst->Named_event_arrays(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Virtual_interface_vars()) {
            auto clone_vec = serializer_->MakeVirtual_interface_varVec();
            inst->Virtual_interface_vars(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Logic_vars()) {
            auto clone_vec = serializer_->MakeLogic_varVec();
            inst->Logic_vars(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Array_vars()) {
            auto clone_vec = serializer_->MakeArray_varVec();
            inst->Array_vars(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Array_var_mems()) {
            auto clone_vec = serializer_->MakeArray_varVec();
            inst->Array_var_mems(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Param_assigns()) {
            auto clone_vec = serializer_->MakeParam_assignVec();
            inst->Param_assigns(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Let_decls()) {
            auto clone_vec = serializer_->MakeLet_declVec();
            inst->Let_decls(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Attributes()) {
            auto clone_vec = serializer_->MakeAttributeVec();
            inst->Attributes(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Parameters()) {
            auto clone_vec = serializer_->MakeAnyVec();
            inst->Parameters(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }
          if (auto vec = defMod->Instance_items()) {
            auto clone_vec = serializer_->MakeAnyVec();
            inst->Instance_items(clone_vec);
            for (auto obj : *vec) {
              clone_vec->push_back(obj->DeepClone(serializer_, this, defMod));
            }
          }


          break;
        }
        default:
          break;
        }
      }
      
    }
  }

  void leaveModule(const module* object, const BaseClass* parent,
                   vpiHandle handle, vpiHandle parentHandle) override {
    const std::string& instName = object->VpiName(); 
    bool flatModule             = (instName == "") && ((object->VpiParent() == 0) ||
                                                       ((object->VpiParent() != 0) && (object->VpiParent()->VpiType() != vpiModule)));
                                  // false when it is a module in a hierachy tree
    if (!flatModule) 
      instStack_.pop();
  }

private:

  // Instance context stack
  std::stack<std::pair<const BaseClass*, std::pair<ComponentMap, ComponentMap>>> instStack_;

  // Flat list of components (modules, udps, interfaces)
  ComponentMap flatComponentMap_;

  Serializer* serializer_;

  bool debug_;
};

};
